# Program za branje datotek s slike diska
# Predpostavlja, da je sliko formatiral CP/M 3 in da ima slika vsaj 256 blokov,
# tako da so sxtevilke njih 16-bitne.

import os

kShraniDatoteke = False  # Ali shranjujemo datoteke
kShraniProsteBloke = True  # Ali shranjujemo proste bloke v datoteke *.dmp
kBrisiProsteBloke = False  # Ali prepisujemo proste bloke z bajti E5
kShraniIzbrisaneDatoteke = True  # Ali shranjujemo tudi izbrisane datoteke

kDolzImenika = 1024  # Koliko vnosov je v imeniku
kDolzBloka = 4096  # Koliko bajtov je dolg en blok (alokacijska enota)
kStRezervBlokov = 2  # Koliko blokov je pred imenikom (imenik se zacxne na #0)

if os.name == "mac":
	kZacRelPoti = ":"
	kSepPoti = ":"
else:
	kZacRelPoti = ""
	kSepPoti = "/"

class tyDatoteka:
	def __init__(self):
		self.Extenti = {}  # Kljucxi so integerji; elementi so arrayi integerjev
		self.StRecordov = 0

def VrniIme(Vnos):
	# Vzame imenisxki vnos in vrne ime ciljne datoteke (npr. "CCP     COM" na
	# uporabnisxkem obmocxju 0 postane ":USER0:CCP.COM" ali podobno).

	def VrniZnak(i):
		return chr(ord(Vnos[i]) & 0x7F)

	Ime = kZacRelPoti
	if ord(Vnos[0]) == 0xE5:
		Ime += "E5"
	else:
		Ime += "USER" + str(ord(Vnos[0]) & 0xF)
	Ime += kSepPoti

	for i in xrange(1, 9):
		Znak = VrniZnak(i)
		if Znak == " ":
			break
		Ime += Znak

	for i in xrange(9, 12):
		Znak = VrniZnak(i)
		if Znak == " ":
			break
		if "." not in Ime:
			Ime += "."
		Ime += Znak

	return Ime

def NiPrazenBlok(iBlok, B):
	# Vzame sxtevilko bloka in blok in vrne True, cxe ni poln znakov E5. Izpisxe
	# tudi kratek komentar o njegovi vsebini.

	Nicvreden = True
	StASCII = 0

	for C in B:
		if C != chr(0xE5):
			Nicvreden = False
			if (ord(C) >= 32 and ord(C) <= 126) or ord(C) in [9, 10, 12, 13, 26]:
				StASCII += 1

	if "Copyright (C) 1985 BORLAND Inc" in B:
		print(str(iBlok) + " je Turbo Pascal program")
	elif "Program aborted" in B:
		print(str(iBlok) + " je del Turbo Pascal programa")
	elif "BRUN.COM not found" in B:
		print(str(iBlok) + " je konec Microsoft BASIC programa")
	elif "COBOL Executor not found" in B:
		print(str(iBlok) + " je del Microsoft COBOL programa")
	elif StASCII >= kDolzBloka * 3 / 4:
		if "begin" in B and "end" in B:
			print(str(iBlok) + " je Pascal koda")
		elif ("CALL" in B and "USING" in B) or "PIC" in B:
			print(str(iBlok) + " je COBOL koda")
		else:
			print(str(iBlok) + " je ASCII besedilna datoteka")
	elif Nicvreden:
		print(str(iBlok) + " je prazen")
	else:
		print(str(iBlok) + " je ???")

	return not Nicvreden

def UstvariImenik(Ime):
	# Ustvari imenik (v trenutnem imeniku), cxe ne obstaja.

	Ime = kZacRelPoti + Ime

	if not os.path.exists(Ime):
		os.mkdir(Ime)

Slika = open(kZacRelPoti + "TrdiDisk.img", "r+b")

# Preverimo, ali je sxtevilo blokov celo in inicializiramo zemljevid zasedenosti.
# Pri tem ne uposxtevamo rezerviranih blokov.
Slika.seek(0, 2)
DolzSlike = Slika.tell() - kStRezervBlokov * kDolzBloka
if DolzSlike % kDolzBloka != 0:
	raise "Sxtevilo blokov ni celo; nepopolna slika?"
StBlokov = DolzSlike / kDolzBloka
Bloki = [False] * StBlokov  # Zemljevid zasedenosti blokov
for iBlok in xrange((kDolzImenika * 32) / kDolzBloka):  # Imenik je zaseden
	Bloki[iBlok] = True

if kShraniDatoteke:
	# Ustvarimo imenike, v katere bomo polagali datoteke s slike.
	if kShraniIzbrisaneDatoteke:
		UstvariImenik("E5")
	for i in xrange(0, 16):
		UstvariImenik("USER" + str(i))

# Beremo imenik in si zapomnimo, katere bloke zasedajo katere datoteke.
Slika.seek(kDolzBloka * kStRezervBlokov, 0)
Datoteke = {}
for i in xrange(0, kDolzImenika):
	Vnos = Slika.read(32)
	Vrsta = ord(Vnos[0])

	if Vrsta & 0xF0 == 0 or (kShraniIzbrisaneDatoteke and Vrsta == 0xE5):
		# Datoteka, po mozxnosti izbrisana
		Ime = VrniIme(Vnos)
		if Vrsta == 0xE5:
			print("Izbrisana datoteka: " + Ime)
		else:
			print("Datoteka: " + Ime)

		try:
			Dat = Datoteke[Ime]
		except KeyError:
			Dat = tyDatoteka()
			Datoteke[Ime] = Dat

		StExtenta = ord(Vnos[12]) * 32 + ord(Vnos[14])
		if ord(Vnos[13]) != 0:
			print("^ LRBC ni 0!")
		Dat.StRecordov += ord(Vnos[15])
		Extent = []

		for i in xrange(16, 32, 2):
			iBlok = ord(Vnos[i]) + ord(Vnos[i + 1]) * 256
			if iBlok != 0:
				if iBlok >= StBlokov:
					print(
						"Prevelika sxtevilka bloka: " + str(iBlok) + " >= " + str(StBlokov)
					)
				else:
					if Bloki[iBlok]:
						print("^ Blok " + str(iBlok) + " zaseda zxe druga datoteka!")
					else:
						Bloki[iBlok] = True
					Extent.append(iBlok)

		Dat.Extenti[StExtenta] = Extent
	elif Vrsta == 0x20:
		# Label
		print("Oznaka diska: " + VrniIme(Vnos))
	elif Vrsta == 0x21:
		# Datestamp
		# TODO
		None
	elif Vrsta & 0xF0 == 0x10:
		# Geslo
		print("Datoteka " + VrniIme(Vnos) + " ima geslo; to ni pomembno.")
	else:
		print("Neznana vrsta vnosa: " + hex(ord(Vnos[0])))

# Zdaj se sprehodimo cxez datoteke, njihove extente, njihove bloke, in njihove
# recorde in jih zapisxemo.
# TODO: trenutno zapisujemo cele bloke in sploh ne uposxtevamo sxtevila blokov.
# Datoteke so zato najvecx 1 blok minus 128 bajtov daljsxe, kot bi morale biti.
if kShraniDatoteke:
	for Ime in Datoteke:
		Dat = Datoteke[Ime]
		Besedilna = \
			Ime.endswith(".TXT") or \
			Ime.endswith(".PAS") or \
			Ime.endswith(".SUB") or \
			Ime.endswith(".INC") or \
			Ime.endswith(".FOR") or \
			Ime.endswith(".MAC") or \
			Ime.endswith(".CMD") or \
			Ime.endswith(".CGM")
		Koncana = False
		Izhod = open(Ime, "wb")
		print("Zapisujem " + Ime + ", " + str(Dat.StRecordov) + " recordov")
		for Extent in Dat.Extenti:
			#print("Extent " + str(Extent))
			for iBlok in Dat.Extenti[Extent]:
				#print(iBlok)
				if not Koncana:
					Slika.seek((iBlok + kStRezervBlokov) * kDolzBloka, 0)
					VsebinaBloka = Slika.read(kDolzBloka)
					if Besedilna:
						if chr(26) in VsebinaBloka:
							VsebinaBloka = VsebinaBloka[:VsebinaBloka.index(chr(26))]
							Koncana = True
					Izhod.write(VsebinaBloka)
		#Izhod.truncate(Dat.StRecordov * 128)  # TO NI V REDU!!!
		Izhod.close()

if kShraniProsteBloke or kBrisiProsteBloke:
	for iBlok in xrange(0, StBlokov):
		if not Bloki[iBlok]:
			Slika.seek((iBlok + kStRezervBlokov) * kDolzBloka, 0)
			if kShraniProsteBloke:
				Temp = Slika.read(kDolzBloka)
				if NiPrazenBlok(iBlok, Temp):
					Izhod = open("Blok" + str(iBlok) + ".dmp", "wb")
					Izhod.write(Temp)
					Izhod.close()
			if kBrisiProsteBloke:
				# Brisxemo proste bloke.
				Slika.write(chr(0xE5) * kDolzBloka)

Slika.close()